/*
 * CinderCones.java
 *
 * Created on April 8, 2005, 11:46 AM
 */

/**
 *
 * @author runyans
 */
import java.io.*;
import java.util.*;
import java.awt.geom.*;

public class CinderCones {
    Vector<Vector<Integer>> freq;
    Hashtable <Float, Point2D.Double> mypoints;// = new Hashtable();
    Vector <Float> keys;
    //Vector distances;// = new Vector(), distances = new Vector();
    File[] allDataFiles;
    BufferedReader br;
    File f1;// = new File(args[1]);
    String path, actual_data_file, output_dir, extention;
    int reps;
    boolean test = false;
    public void debug(String s){
        if(test)
            System.out.println(s);
    }
    public CinderCones(String directory, String file){
        reps = 0;
        actual_data_file = file;
        f1 = new File(directory);
        allDataFiles = f1.listFiles();
        path = directory;//file.substring(0, file.lastIndexOf('\\')+1);
        freq = new Vector<Vector<Integer>>();
        for(int i =0; i< 36; i++)//36 total azimuth bins
            freq.add(new Vector<Integer>());
        reset();
        try{
            File f = makeOutputDir();
            output_dir = f.toString();
            debug(output_dir);
        }catch(IOException e){
            System.err.println("Cannot create output "+e);
        }
        runDataSet();
    }
    public void reset(){
        mypoints = new Hashtable<Float, Point2D.Double>();
        keys = new Vector<Float>();
        //distances = new Vector();
    }
    public void runDataSet(){
        String filein = "", fileout = "";
         for (int i = 0; i < allDataFiles.length; i++){
            //for each file in test data directory
            if(allDataFiles[i].isFile()){
                filein = (allDataFiles[i].toString()).substring((allDataFiles[i].toString()).lastIndexOf('\\')+1);
                extention = filein.substring(filein.indexOf('.'));
                fileout = filein.substring(0, filein.indexOf('.'))+"_out"+extention;//filename to write data out to 
            
                try{
                    br = new BufferedReader(new FileReader(allDataFiles[i]));               
                }catch(IOException e){
                    System.err.println("Exception reading from file"+filein+": "+e);
                    System.exit(0);
                }
                try{
                    recordPts();//fill mypoints and keys with data from current file
                }catch(Exception e){
                    System.err.println("Exception when filling array "+e);
                    System.exit(0);
                }
                writeOut(output_dir+"\\"+fileout);//write output file with calculated distances & azimuths and add distances to freq
                reps++;
                reset();//reset mypoints, keys and distances variables
                frequency(filein.substring(0,filein.indexOf('.'))+"_frequency"+extention);
            }
         }
        debug("before summary");
        frequencySummary();
        debug("after summary");
         calculateActualData();
         debug(freq.toString());
    }
    public void recordPts() throws Exception{
        String line=null;
        StringTokenizer st;
        double x,y;
        Float id;
        int count = 0, size;
        try{
            line = br.readLine();//header line
            line = br.readLine();//first line of data
        }catch(IOException e){
             System.err.println("Exception when reading from file :"+e);
             System.exit(0);
        }
        while (line != null){
            st = new StringTokenizer(line, ",");
            if(st.countTokens() != 3){
                throw new Exception("Invalid input from file:"+line);
            }
            else{
                try{
                x = new Double(st.nextToken()).doubleValue();
                y = new Double(st.nextToken()).doubleValue();
                id = new Float(st.nextToken());
                }catch (NumberFormatException nfe){
                    throw new Exception("Invalid format from file: "+nfe);
                }
                //add new MyPoint to vector.  Remember to cast to MyPoint when retrieving from Vector
                //ie: MyPoint pt = (MyPoint)mypoints.elementAt(index);
                //mypoints.add(new MyPoint(x,y,id)); 
                mypoints.put(id, new Point2D.Double(x,y));
                keys.add(id);
            }
            try{
                line = br.readLine();
            }catch(IOException e){
                System.err.println("Exception when reading from file :"+e);
            }
        }
    }
    public void writeOut(String fileout){
        PrintWriter out;
        Point2D.Double pt1, pt2;
        Float id1, id2;
        int count=0;
        int [] azCnt = new int[36];
        for(int i = 0; i < azCnt.length; i++)
            azCnt[i] = 0;
        try{
            out = new PrintWriter(new BufferedWriter(new FileWriter(fileout)));
            out.println("Line Segment ID, Distance (km), Azimuth");
            int start = 1;
            for (int i = 0; i < keys.size(); i++){
                id1 = keys.get(i);
                pt1 = mypoints.get(id1);
                for(int j = start; j < keys.size(); j++){
                    //calc distance
                    id2 = keys.get(j);
                    pt2 = mypoints.get(id2);
                    if(!pt1.equals(pt2)){
                        double dist = pt1.distance(pt2);//calculateDistance(id1, id2);
                        double az = calculateAzimuth(id1, id2);
                        azCnt[(new Double(17+Math.ceil(az/5.0))).intValue()]++;
                        String key1, key2;
                        dist = (Math.round(dist/10))/100.0;//m to km with 2 values after decimal
                        az = (Math.round(az*100.0))/100.0;
                        if(((Float)keys.get(i)).floatValue() == ((Float)keys.get(i)).intValue())
                            key1 = ""+((Float)keys.get(i)).intValue();
                        else
                            key1 = ""+((Float)keys.get(i)).floatValue();
                        if(((Float)keys.get(j)).floatValue() == ((Float)keys.get(j)).intValue())
                            key2 = ""+((Float)keys.get(j)).intValue();
                        else
                            key2 = ""+((Float)keys.get(j)).floatValue();
                        out.println(key1+" : "+key2+", "+dist+", "+az);
                        //out.println();
                        count++;
                        if(count > 20000){
                            out.flush();
                        }
                    }
                }
                start++;
            }
            out.flush();
            out.close();
            updateFreq(azCnt);
        }catch(IOException e){
            System.err.println("Cannot open file "+fileout+" to write: "+e);
        }
    }
    public void updateFreq(int [] azimuths){
        for (int i = 0; i < azimuths.length; i++){
            Vector<Integer> v = freq.get(i);//get current count for this azimuth bin
            v.add(new Integer(azimuths[i]));//add count for file
            freq.set(i, v);//replace freq for this bin with current data added.
        }
    }
    public double calculateAzimuth(Float id1, Float id2){
        Point2D.Double pt1, pt2;
        pt1 = (Point2D.Double)mypoints.get(id1);
        pt2 = (Point2D.Double)mypoints.get(id2);
        if(pt1.getX() == pt2.getX()) return 0.0;
        else if(pt1.getY() == pt2.getY()) return 90.0;
        else{ 
            double sign = 1.0, opp, dis, oppoverdis, degrees;
            if((pt1.getY()-pt2.getY())/(pt1.getX()-pt2.getX())<0) sign = -1.0;
            opp = Math.abs(pt1.getY()-pt2.getY());
            dis = pt1.distance(pt2);
            oppoverdis = (Math.abs(pt1.getY()-pt2.getY()))/pt1.distance(pt2);
            degrees = 90-(Math.toDegrees(Math.asin(oppoverdis)));
            return degrees*sign;
        }
    }
    public double[] calculateMeanAndSD (Vector<Integer> azCnt)
    {
        double meanSum = 0.0, sigmaSum = 0.0;
        double mean = 0.0;
        double sigma = 0.0; // standard deviation;
        double max = 0.0, min = 0.0;

        Enumeration enumeration = azCnt.elements();
        while( enumeration.hasMoreElements() )
        {
            Integer dd = (Integer) enumeration.nextElement();
            max = Math.max(dd.doubleValue(), max);
            if(meanSum == 0.0) min = max;
            min = Math.min(dd.doubleValue(), min);
            meanSum += dd.doubleValue();
        }

        mean = meanSum / azCnt.size();
        
        enumeration = azCnt.elements();
        while(enumeration.hasMoreElements()){
            Integer dd = (Integer) enumeration.nextElement();
            sigmaSum += Math.pow(dd.doubleValue(), 2) - (2*mean*dd.doubleValue()) + Math.pow(mean, 2);
        }
        sigmaSum = sigmaSum / (azCnt.size() - 1); // divide by n-1
        sigma = Math.sqrt(sigmaSum);
        
        double[] returnArray = { max, min, mean, sigma};
        debug(""+returnArray[0]+" "+returnArray[1]+" "+returnArray[2]+" "+returnArray[3]);
        return returnArray;
    }
    public void frequencySummary(){
        PrintWriter out;
        int high, low;
        try{
            out = new PrintWriter(new BufferedWriter(new FileWriter(output_dir+"\\random_frequency_summary"+extention)));
            out.println("Azimuth Bin, Max, Min, Mean, Standard Deviation, Number of Files");
            //out.println();
            out.flush();
            for(int i = 0; i<freq.size(); i++){
                if(i < 17){
                    high = (17-i)*(-5);
                    low = high - 5;
                }else if(i > 17){
                    high = (i-17)*5;
                    low = high - 5;
                }else{//i=17
                    high = 0;
                    low = -5;
                }
                double[] meanAndSD = calculateMeanAndSD(freq.get(i));
                //debug("Max:"+meanAndSD[0]);
                //debug("Min:"+meanAndSD[1]);
                //meanAndSD[0] = max, [1] = min, [2] = mean, [3] = sd
                String toPrint = "("+low+" : "+high+"] , ";
                toPrint += meanAndSD[0]/1 +", "+meanAndSD[1]/1+", ";
                toPrint += (Math.round(meanAndSD[2] * 100.0))/100.0+", ";
                toPrint += (Math.round(meanAndSD[3] * 100.0))/100.0 +", "+reps;
                out.println(toPrint);
            }
            out.close();
        }catch(IOException e){
                System.err.println("Exception writing to file "+output_dir+"\\random_frequency_summary"+extention);
                System.exit(0);
        }
    }
    public void frequency(String fileout){
      PrintWriter out;
      int[] cnt = new int[freq.size()];
      int high, low, sum = 0;
      for(int i = 0; i < freq.size(); i++){
        Vector <Integer> v = freq.get(i);
        cnt[i] = v.lastElement().intValue();
        sum += cnt[i];
      }
      try{
        out = new PrintWriter(new BufferedWriter(new FileWriter(output_dir+"\\"+fileout)));
        out.println("Azimuth Bin, Raw Count, Percentage");
        //out.println();
        out.flush();
        for(int i = 0; i < cnt.length; i++){
            if(i < 17){
                high = (17-i)*(-5);
                low = high - 5;
            }else if(i > 17){
                high = (i-17)*5;
                low = high - 5;
            }else{//i=17
                high = 0;
                low = -5;
            }
            double perc = (new Double(cnt[i])).doubleValue() / (new Double(sum)).doubleValue();
            perc = (Math.round(10000.0 * perc))/100.0;
            out.println("("+low+" : "+high+"] , "+cnt[i]+", "+perc+"%");
        }
        out.flush();
        out.close();
      }catch(IOException e){
              System.err.println("Exception writing to file "+output_dir+"\\actual_frequency.csv");
              System.exit(0);
      }
    }
    public File makeOutputDir() throws IOException{
        int dirNum = 0;
        File dir;
        while(true){
            dir = new File(path, "output."+System.currentTimeMillis());
            if(dir.mkdirs()){
                return dir;
            }
            if(!dir.exists()){
                //didn't fail because it already exists, must be an access error.
                throw new IOException("No permission to create a directory "+dir.getAbsolutePath());
            }
            dirNum++;
            if(dirNum >= 25){
                throw new IOException ("Failed to create tmp directory named in the "+
                                        "form \""+dir.getAbsolutePath() +
                                        "\" after "+dirNum+" trys.");
            }
        }
    }
    public void calculateActualData(){
        String filein = actual_data_file.substring(actual_data_file.lastIndexOf('\\')+1);
        String extention = filein.substring(filein.indexOf('.'));
        String fileout = filein.substring(0, filein.indexOf('.'))+"_out"+extention;
        PrintWriter out;
        int high, low, sum = 0;
        int[] cnt = new int[freq.size()];
//        for(int i = 0; i < freq.size(); i++){
//            Vector <Integer> v = freq.get(i);
//            cnt[i] = v.lastElement().intValue();
//            sum += cnt[i];
//        }
        try{
            br = new BufferedReader(new FileReader(actual_data_file));               
        }catch(IOException e){
            System.err.println("Exception reading from file: "+e);
            System.exit(0);
        }
        try{
            recordPts();//fill mypoints and keys with data from current file
        }catch(Exception e){
            System.err.println("Exception when filling array "+e);
            System.exit(0);
        }
        writeOut(output_dir+"\\"+fileout);//write output file with calculated distances & azimuths and add distances to freq
        for(int i = 0; i < freq.size(); i++){
            Vector <Integer> v = freq.get(i);
            cnt[i] = v.lastElement().intValue();
            sum += cnt[i];
        }
        try{
            out = new PrintWriter(new BufferedWriter(new FileWriter(output_dir+"\\actual_frequency"+extention)));
            out.println("Azimuth Bin, Raw Count, Percentage");
            //out.println();
            out.flush();
            for(int i = 0; i < cnt.length; i++){
                if(i < 17){
                    high = (17-i)*(-5);
                    low = high - 5;
                }else if(i > 17){
                    high = (i-17)*5;
                    low = high - 5;
                }else{//i=17
                    high = 0;
                    low = -5;
                }
                double perc = (new Double(cnt[i])).doubleValue() / (new Double(sum)).doubleValue();
                perc = (Math.round(10000.0 * perc))/100.0;
                out.println("("+low+" : "+high+"] , "+cnt[i]+", "+perc+"%");
            }
            out.flush();
            out.close();
        }catch(IOException e){
                System.err.println("Exception writing to file "+output_dir+"\\actual_frequency.csv");
                System.exit(0);
        }
    }
    public static void main(String[] args){
        String directory = "", data_file = "";
        CinderCones mycones;
        for (int i = 0; i < args.length; i++){
            if(args[i].equals("-d"))
                directory = args[i+1];
            else if(args[i].equals("-f"))
                data_file = args[i+1];
        }
        if(directory != "" && data_file != "")
            mycones = new CinderCones(directory, data_file);
        else{
            System.out.println("Invalid usage");
            System.out.println("java CinderCones -d full_directory_path -f full_file_path");
            if(args.length > 4){
                System.out.println("Too many arguments!");
                System.out.println("If your file name or path includes spaces, be sure to put them in quotes: ");
                System.out.println("\tjava CinderCones -d \"C:\\Test\\TestData\" -f \"C:\\Test\\actualdata.txt\"");
            }
            System.exit(0);
        }
    }
}